// CartoFormView.cpp : implementation of the CCartoFormView class
//

#include "stdafx.h"
#include "Carto.h"

#include "CartoDoc.h"
#include "CartoFormView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCartoFormView

IMPLEMENT_DYNCREATE(CCartoFormView, CFormView)

BEGIN_MESSAGE_MAP(CCartoFormView, CFormView)
	//{{AFX_MSG_MAP(CCartoFormView)
	ON_BN_CLICKED(IDC_BNLOAD, OnLoadCoord)
	ON_BN_CLICKED(IDC_BNGO, OnGo)
	ON_BN_CLICKED(IDC_BNSTOP, OnStop)
	ON_BN_CLICKED(IDC_BNSAVE, OnSavePS)
	ON_WM_HSCROLL()
	ON_BN_CLICKED(IDC_BNRESUME, OnResume)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCartoFormView construction/destruction

CCartoFormView::CCartoFormView()
	: CFormView(CCartoFormView::IDD)
{
	//{{AFX_DATA_INIT(CCartoFormView)
	m_TotCount = 0;
	m_nCurFitness = 0;
	m_nCurFitPct = 0;
	m_nBestGen = 0;
	m_nBestFitPct = 0;
	m_nBestFitness = 0;
	m_nCurGen = 0;
	m_nMutations = 0;
	//}}AFX_DATA_INIT
	// TODO: add construction code here

}

CCartoFormView::~CCartoFormView()
{
}

void CCartoFormView::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CCartoFormView)
	DDX_Control(pDX, IDC_MUTATIONS, m_MutationSlider);
	DDX_Control(pDX, IDC_BNSAVE, m_SaveButton);
	DDX_Control(pDX, IDC_BNSTOP, m_StopButton);
	DDX_Control(pDX, IDC_GENDISPLAY, m_GenGraph);
	DDX_Control(pDX, IDC_BNRESUME, m_ResumeButton);
	DDX_Control(pDX, IDC_BNLOAD, m_LoadButton);
	DDX_Control(pDX, IDC_BNGO, m_GoButton);
	DDX_Text(pDX, IDC_TOTCOUNT, m_TotCount);
	DDX_Text(pDX, IDC_CLRCOUNT, m_nCurFitness);
	DDX_Text(pDX, IDC_CLRPCT, m_nCurFitPct);
	DDX_Text(pDX, IDC_BESTGENCOUNT, m_nBestGen);
	DDX_Text(pDX, IDC_BESTCLRPCT, m_nBestFitPct);
	DDX_Text(pDX, IDC_BESTCLRCOUNT, m_nBestFitness);
	DDX_Text(pDX, IDC_GENCOUNT, m_nCurGen);
	DDX_Text(pDX, IDC_MUTATIONCNT, m_nMutations);
	//}}AFX_DATA_MAP
}

BOOL CCartoFormView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CFormView::PreCreateWindow(cs);
}

void CCartoFormView::OnInitialUpdate()
{
	CFormView::OnInitialUpdate();
	GetParentFrame()->RecalcLayout();
	ResizeParentToFit();

	m_GenGraph.Init(1);		 // Show the graph
}

/////////////////////////////////////////////////////////////////////////////
// CCartoFormView diagnostics

#ifdef _DEBUG
void CCartoFormView::AssertValid() const
{
	CFormView::AssertValid();
}

void CCartoFormView::Dump(CDumpContext& dc) const
{
	CFormView::Dump(dc);
}

CCartoDoc* CCartoFormView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CCartoDoc)));
	return (CCartoDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCartoFormView message handlers

void CCartoFormView::OnLoadCoord() 
{
	CFileDialog fdlg(TRUE);
	if (fdlg.DoModal() == IDOK)
	{
		// Read in the new points and give them to the Doc:
		GetDocument()->ClearPts();
		CString pathname = fdlg.GetPathName();
		CStdioFile coordFile(pathname, CFile::modeRead);
		CString coordString;
		CPoint pt;
		int ptCnt = 0;
		int x, y;
		while (coordFile.ReadString(coordString))
		{
			// ignore comments:
			if ((coordString.Find("//") == 0) || (coordString.Find("#") == 0))
				continue;
			// ignore blank lines:
			if (coordString.IsEmpty())
				continue;

			// Trim off leading whitespace:
			coordString.TrimLeft();
			// Get first coordinate:
			x = atoi(coordString.SpanIncluding("0123456789"));

			// Skip a space:
			coordString = coordString.Mid(coordString.Find(' ')+1);
			// Get second coordinate:
			y = atoi(coordString.SpanIncluding("0123456789"));

			// Now that we have the coordinates of a new point
			// Update the underlying document:
			ptCnt++;
			pt.x = x;
			pt.y = y;
			GetDocument()->AddPt(pt);
		}
		// Show the new map:
		GetDocument()->Init();
		GetDocument()->UpdateAllViews(this);

		// Update the interface state:
		m_GenGraph.Init(ptCnt); // Scale the graph
		m_nMutations = 100;
		m_MutationSlider.EnableWindow(TRUE);
		m_MutationSlider.SetRange(0, 50 * ptCnt);
		m_MutationSlider.SetPos(m_nMutations);

		m_TotCount = ptCnt;
		UpdateData(FALSE);
		m_LoadButton.SetButtonStyle(BS_PUSHBUTTON);
		m_ResumeButton.EnableWindow(FALSE);
		m_GoButton.EnableWindow(TRUE);
		m_GoButton.SetButtonStyle(BS_DEFPUSHBUTTON);
		m_GoButton.SetFocus();
	}
}

void CCartoFormView::OnGo() 
{
	// Change the text on the Go button on the first click
	// so the user doesn't confuse it with a resume:
	m_GoButton.SetWindowText("Start A&gain");
	// Set the button states:
	m_LoadButton.EnableWindow(FALSE);
	m_SaveButton.EnableWindow(FALSE);
	m_GoButton.EnableWindow(FALSE);
	m_GoButton.SetButtonStyle(BS_PUSHBUTTON);
	m_StopButton.EnableWindow(TRUE);
	m_StopButton.SetFocus();
	m_StopButton.SetButtonStyle(BS_DEFPUSHBUTTON);
	m_ResumeButton.EnableWindow(FALSE);
	m_ResumeButton.SetButtonStyle(BS_PUSHBUTTON);

	// Start (or restart) the Generations:
	// Reset everything to 0:
	CCartoDoc* pDoc = GetDocument();
	pDoc->Init();		// The Doc's m_PointArray is still set
	m_GenGraph.AddFitness(0, 0);

	// Initialize variables (and UI):
	m_nCurFitness = 0;
	m_nCurFitPct = 0;
	m_nCurGen = 0;
	m_nBestFitness = 0;
	m_nBestFitPct = 0;
	m_nBestGen = 0;
	UpdateData(FALSE);

	CString cs;
	// god is immortal:
	m_bRunLoop = TRUE;
	while (m_bRunLoop)
	{
		// Let there be electromagnetic radiation:
		pDoc->m_Gen.SetMutationRate(m_nMutations);

		// Be fruitful and multiply:
		m_nCurFitness = pDoc->m_Gen++;
		m_nCurGen++;
		m_nCurFitPct = (int)((float)m_nCurFitness / pDoc->m_PtCnt * 100);

		// And god looked upon his creatures and saw they were good:
		if (m_nCurFitness >= m_nBestFitness)
		{
			m_nBestFitness = m_nCurFitness;

			// Let all see the majesty of my creation:
			pDoc->UpdateAllViews(this);

			m_nBestGen = m_nCurGen;
			m_nBestFitPct = m_nCurFitPct;
			cs.Format("%d%%", m_nBestFitPct);
			GetParentFrame()->SetWindowText(cs + "- Carto");
			if (m_nCurFitness == pDoc->m_PtCnt)
			{
				// Perfection. The universe can end now.
				m_bRunLoop = FALSE;
				m_LoadButton.EnableWindow(TRUE);
				m_SaveButton.EnableWindow(TRUE);
				m_GoButton.EnableWindow(TRUE);
				m_GoButton.SetFocus();
				m_GoButton.SetButtonStyle(BS_DEFPUSHBUTTON);
				m_StopButton.EnableWindow(FALSE);
				m_StopButton.SetButtonStyle(BS_PUSHBUTTON);
				m_ResumeButton.EnableWindow(FALSE);
			}
		}
		m_GenGraph.AddFitness(m_nCurGen, m_nCurFitness);
		UpdateData(FALSE);

		// Even god must obey the user:
		if(!PeekAndPump())
			return;
	}
}

void CCartoFormView::OnStop() 
{
	// Set the button states:
	m_LoadButton.EnableWindow(TRUE);
	m_SaveButton.EnableWindow(TRUE);
	m_GoButton.EnableWindow(TRUE);
	m_StopButton.EnableWindow(FALSE);
	m_StopButton.SetButtonStyle(BS_PUSHBUTTON);
	m_ResumeButton.EnableWindow(TRUE);
	m_ResumeButton.SetFocus();
	m_ResumeButton.SetButtonStyle(BS_DEFPUSHBUTTON);

	// Stop the generations:
	m_bRunLoop = FALSE;
}

BOOL CCartoFormView::PeekAndPump()
{
	MSG msg;
	while(::PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE))
	{
		if(!AfxGetApp()->PumpMessage())
		{
			::PostQuitMessage(0);
			return FALSE;
		}
	}
	LONG lIdle = 0;
	while(AfxGetApp()->OnIdle(lIdle++));
	return TRUE;
}

void CCartoFormView::OnSavePS() 
{
	// Saves the Best Ever (displayed in the map view)
	// as a Postscript file.

	// First things first:
	CFileDialog fdlg(FALSE, "ps", "results",
		OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
		"Postscript Files (*.ps)|*.ps|All Files (*.*)|*.*||",
		NULL);
	if (fdlg.DoModal() == IDOK)
	{
		// Read in the new points and give them to the Doc:
		CString pathname = fdlg.GetPathName();
		CStdioFile ResultsFile(pathname,
			CFile::modeCreate | CFile::modeWrite);

		// Put in the header.  This is a little crude:
		ResultsFile.WriteString("%! PostScript map\n\n");

		ResultsFile.WriteString("/draw_border\n{\n  newpath\n");
		ResultsFile.WriteString("  36 36 moveto\n  0 720 rlineto\n");
		ResultsFile.WriteString("  540 0 rlineto\n  0 -720 rlineto\n");
		ResultsFile.WriteString("  closepath\n  2 setlinewidth\n");
		ResultsFile.WriteString("  stroke\n} def\n\n");

		ResultsFile.WriteString("/draw_town\n{\n  3 0 360 arc\n");
		ResultsFile.WriteString("  closepath\n  fill\n} def\n\n");

		ResultsFile.WriteString("/UR_label\n{\n  moveto\n");
		ResultsFile.WriteString("  0 12 rlineto\n  60 0 rlineto\n");
		ResultsFile.WriteString("  0 -12 rlineto\n  closepath\n");
		ResultsFile.WriteString("  1 setlinewidth\n  {stroke} ");
		ResultsFile.WriteString("{gsave 0.6 setgray fill grestore ");
		ResultsFile.WriteString("0.0 setgray stroke} ifelse\n} def\n\n");

		ResultsFile.WriteString("/UL_label\n{\n  moveto\n");
		ResultsFile.WriteString("  0 12 rlineto\n  -60 0 rlineto\n");
		ResultsFile.WriteString("  0 -12 rlineto\n  closepath\n");
		ResultsFile.WriteString("  1 setlinewidth\n  {stroke} ");
		ResultsFile.WriteString("{gsave 0.6 setgray fill grestore ");
		ResultsFile.WriteString("0.0 setgray stroke} ifelse\n} def\n\n");

		ResultsFile.WriteString("/LR_label\n{\n  moveto\n");
		ResultsFile.WriteString("  0 -12 rlineto\n  60 0 rlineto\n");
		ResultsFile.WriteString("  0 12 rlineto\n  closepath\n");
		ResultsFile.WriteString("  1 setlinewidth\n  {stroke} ");
		ResultsFile.WriteString("{gsave 0.6 setgray fill grestore ");
		ResultsFile.WriteString("0.0 setgray stroke} ifelse\n} def\n\n");

		ResultsFile.WriteString("/LL_label\n{\n  moveto\n");
		ResultsFile.WriteString("  0 -12 rlineto\n  -60 0 rlineto\n");
		ResultsFile.WriteString("  0 12 rlineto\n  closepath\n");
		ResultsFile.WriteString("  1 setlinewidth\n  {stroke} ");
		ResultsFile.WriteString("{gsave 0.6 setgray fill grestore ");
		ResultsFile.WriteString("0.0 setgray stroke} ifelse\n} def\n\n");

		// Now put in the instructions to draw the best map:
		ResultsFile.WriteString("\ndraw_border\n");
		CCartoDoc* pDoc = GetDocument();
		CString outString, lblString;
		for (int i=0; i<pDoc->m_PtCnt; i++)
		{
			switch (pDoc->m_Gen.bestEver.m_Genes[i])
			{
			  case UR:
				lblString = "UR_label";
				break;
			  case UL:
				lblString = "UL_label";
				break;
			  case LR:
				lblString = "LR_label";
				break;
			  case LL:
				lblString = "LL_label";
				break;
			}
			outString.Format("%s %d %d %s\n",
				(pDoc->m_Gen.bestEver.m_Overlaps[i] == TRUE)?"false":"true",
				pDoc->m_Gen.bestEver.m_Points[i].x,
				pDoc->m_Gen.bestEver.m_Points[i].y,
				lblString);
			ResultsFile.WriteString(outString);
			outString.Format("%d %d draw_town\n",
				pDoc->m_Gen.bestEver.m_Points[i].x,
				pDoc->m_Gen.bestEver.m_Points[i].y);
			ResultsFile.WriteString(outString);
		}
		ResultsFile.WriteString("showpage\n");
	}
}

void CCartoFormView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// The Mutation Rate got changed:
	if (nSBCode == TB_THUMBPOSITION)
		m_nMutations = nPos;
	UpdateData(FALSE);

	CFormView::OnHScroll(nSBCode, nPos, pScrollBar);
}

void CCartoFormView::OnResume() 
{
	// Set the button states:
	m_LoadButton.EnableWindow(FALSE);
	m_SaveButton.EnableWindow(FALSE);
	m_GoButton.EnableWindow(FALSE);
	m_GoButton.SetButtonStyle(BS_PUSHBUTTON);
	m_StopButton.EnableWindow(TRUE);
	m_StopButton.SetFocus();
	m_StopButton.SetButtonStyle(BS_DEFPUSHBUTTON);
	m_ResumeButton.EnableWindow(FALSE);
	m_ResumeButton.SetButtonStyle(BS_PUSHBUTTON);

	// Continue the Generations without reinitializing
	CCartoDoc* pDoc = GetDocument();

	CString cs;
	// god is immortal:
	m_bRunLoop = TRUE;
	while (m_bRunLoop)
	{
		// Let there be electromagnetic radiation:
		pDoc->m_Gen.SetMutationRate(m_nMutations);

		// Be fruitful and multiply:
		m_nCurFitness = pDoc->m_Gen++;
		m_nCurGen++;
		m_nCurFitPct = (int)((float)m_nCurFitness / pDoc->m_PtCnt * 100);

		// And god looked upon his creatures and saw they were good:
		if (m_nCurFitness > m_nBestFitness)
		{
			m_nBestFitness = m_nCurFitness;

			// Let all see the majesty of my creation:
			pDoc->UpdateAllViews(this);

			m_nBestGen = m_nCurGen;
			m_nBestFitPct = m_nCurFitPct;
			cs.Format("%d%%", m_nBestFitPct);
			GetParentFrame()->SetWindowText(cs + "- Carto");
			if (m_nCurFitness == pDoc->m_PtCnt)
			{
				// Perfection. The universe can end now.
				m_bRunLoop = FALSE;
				m_LoadButton.EnableWindow(TRUE);
				m_SaveButton.EnableWindow(TRUE);
				m_GoButton.EnableWindow(TRUE);
				m_GoButton.SetFocus();
				m_GoButton.SetButtonStyle(BS_DEFPUSHBUTTON);
				m_StopButton.EnableWindow(FALSE);
				m_StopButton.SetButtonStyle(BS_PUSHBUTTON);
				m_ResumeButton.EnableWindow(FALSE);
			}
		}
		m_GenGraph.AddFitness(m_nCurGen, m_nCurFitness);
		UpdateData(FALSE);

		// Even god must obey the user:
		if(!PeekAndPump())
			return;
	}
}
