// Layout.cpp : implementation file
//

#include "stdafx.h"
#include "Carto.h"
#include "Layout.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CLayout

CLayout::CLayout()
	: m_randomPos(0, 3), m_randomParent(0, 1)
{
	// Set up the arrays (they stay empty until Init())
	m_Points.SetSize(100);
	m_Points.RemoveAll();
	m_Rects.SetSize(100);
	m_Rects.RemoveAll();
	m_Genes.SetSize(100);
	m_Genes.RemoveAll();
	m_Overlaps.SetSize(100);
	m_Overlaps.RemoveAll();
}

// The copy ctor:
CLayout::CLayout(const CLayout &copy)
{
	for (int i=0; i<GetPtCnt(); i++)
	{
		m_Genes[i] = copy.m_Genes[i];
		m_Overlaps[i] = copy.m_Overlaps[i];
		m_Rects[i] = copy.m_Rects[i];
		m_Points[i] = copy.m_Points[i];
	}
}

CLayout::~CLayout()
{
}


BEGIN_MESSAGE_MAP(CLayout, CStatic)
	//{{AFX_MSG_MAP(CLayout)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void CLayout::Init(const CArray<CPoint, CPoint&>& ptarray)
{
	// Called when the Generation collection is Init'ed;
	// make it a new size and create random genes.
	m_Genes.RemoveAll();
	m_Overlaps.RemoveAll();
	m_Points.RemoveAll();
	m_Rects.RemoveAll();

	// Make the CLayout objects the right size:
	CRect r(0,0,0,0);			// Just some fillers
	BOOL ol = FALSE;
	int g;
	for (int i=0; i<ptarray.GetSize(); i++)
	{
		g = m_randomPos.Pick();
		m_Genes.Add(g);			// Original random genes
		m_Points.Add(ptarray.GetAt(i));
		m_Rects.Add(r);			// Calc'd by DetermineFitness
		m_Overlaps.Add(ol);		// Calc'd by DetermineFitness
	}
}

int CLayout::GetPtCnt()
{
	return m_Points.GetSize();
}

int CLayout::DetermineFitness()
{
	// The Generation collection uses this method to
	// choose parents.  The bigger the better.

	// Calling this routine has these side effects:
	// - Sets the rectangle array which is consulted
	//    by the overlap checker as well as the MapView's
	//    OnDraw.
	// - Sets the overlap array to indicate collisions.

	// This is called a lot, so let's optimize some:
	int ptCnt = GetPtCnt();

	// First, initialize all the Overlaps:
	int olCnt = 0;
	for (int i=0; i<ptCnt; i++)
		m_Overlaps[i] = FALSE;

	// Calculate the rectangles now (loop n times)
	// so they don't have to be done (n^2)/2 times
	// during the check for intersection
	CRect r;
	CPoint pt;
	for (i=0; i<ptCnt; i++)
	{
		// Calculate this rectangle:
		pt = m_Points[i];
		switch (m_Genes[i])
		{
		  case UL:
			r = CRect(pt.x-LBL_WIDTH, pt.y,
						pt.x+1, pt.y+LBL_HEIGHT+1);
			break;
		  case UR:
			r = CRect(pt.x, pt.y,
						pt.x+LBL_WIDTH+1, pt.y+LBL_HEIGHT+1);
			break;
		  case LL:
			r = CRect(pt.x-LBL_WIDTH, pt.y-LBL_HEIGHT,
						pt.x+1, pt.y+1);
			break;
		  case LR:
			r = CRect(pt.x, pt.y-LBL_HEIGHT,
						pt.x+LBL_WIDTH+1, pt.y+1);
			break;
		}
		m_Rects[i] = r;
	}

	// Figure out how many of these rectangles intersect.
	// Check 0 against 1 thru n, record all overlaps.
	// Check 1 against 2 thru n, record all overlaps.
	// :
	// Check n-2 against n-1 thru n, record all overlaps.
	// Check n-1 against n, record all overlaps.
	for (i=0; i<ptCnt-1; i++)
	{
		// Check the ith rectangle against all the others
		// (This is the P part of the NP-complete problem):
		for (int j=i+1; j<ptCnt; j++)
		{
			// OK, now for the moment of truth:
			// r is a dummy here, it is set to a bounding rect
			// which we don't need.  But we want ith & jth to
			// be unchanged--it may be used by MapView's OnDraw
			if (r.IntersectRect(m_Rects[i], m_Rects[j]) != 0)
			{
				m_Overlaps[i] = TRUE;
				m_Overlaps[j] = TRUE;
			}
		}
	}
	// Count the overlaps:
	// Why wasn't a counter keeping track during the checks?
	// Because one rect may intersect many others but should
	// only count as one overlapped rect.
	for (i=0; i<ptCnt; i++)
		if (m_Overlaps[i] == TRUE)
			olCnt++;

	return ptCnt - olCnt;
}

// The (=) assignment:
CLayout& CLayout::operator=(const CLayout &right)
{
	if (&right == this)
		return *this;

	// Since both sides already exist, just copy the contents:
	for (int i=0; i<GetPtCnt(); i++)
	{
		m_Genes[i] = right.m_Genes[i];
		m_Overlaps[i] = right.m_Overlaps[i];
		m_Points[i] = right.m_Points[i];
		m_Rects[i] = right.m_Rects[i];
	}
	
	return *this;
}

// This was originally the overloaded operator*() so a member
// could be set like this:
// GenMember[i] = GenMember[Parent[i]] * GenMember[Parent[j]];
// But that required a construction of a child, an Init of
// the child, a call to the copy ctor, a call to operator=(),
// then a destruction of the child!
// This is done millions of times, let's save some time:
// 
void CLayout::MakeChild(const CLayout &mom, const CLayout &dad)
{
	// OK, let's have sex!
	for (int i=0; i<GetPtCnt(); i++)
	{
		// There are two schools of thought about what to
		// do now.  We could:
		// (1)  Find which of the parents has no overlap
		//      for each point and take the superior genes
		//		If they're both good, flip a coin, if
		//		they're both bad, pick a new random gene.
		//		(This may be faster to reach a solution)
#define GUY_BEATS_NATURE
#ifdef GUY_BEATS_NATURE

		if (mom.m_Overlaps[i] == TRUE)	// Mom's gene is no good
		{
			if (dad.m_Overlaps[i] == TRUE)	// Dad's gene is n.g.
				m_Genes[i] = m_randomPos.Pick();
			else	// Dad's is better than Mom's
				m_Genes[i] = dad.m_Genes[i];
		}
		else	// Mom's gene is good
		{
			if (dad.m_Overlaps[i] == TRUE)	// Dad's gene is n.g.
				m_Genes[i] = mom.m_Genes[i];
			else	// Dad's and Mom's are both OK
			{
				if (m_randomParent.Pick() == 0)
					m_Genes[i] = mom.m_Genes[i];
				else
					m_Genes[i] = dad.m_Genes[i];
			}
		}

#endif
		// OR
		// (2)  Just randomly pick either Mom's or Dad's
		//      genes.  (This is more like Nature's scheme).
#ifndef GUY_BEATS_NATURE

		if (m_randomParent.Pick() == 0)
			m_Genes[i] = mom.m_Genes[i];
		else
			m_Genes[i] = dad.m_Genes[i];

#endif

	}
}
